'use strict';
Object.defineProperty(exports, '__esModule', { value: true });
exports.insertNativeElementByIndex =
  exports.setTrackUpdate =
  exports.finishCommit =
  exports.commit =
  exports.createNativeElement =
    void 0;
const core_1 = require('@dark-engine/core');
const portal_1 = require('../portal');
const events_1 = require('../events');
const constants_1 = require('../constants');
const attrBlackListMap = {
  [core_1.ATTR_KEY]: true,
  [core_1.ATTR_REF]: true,
  [core_1.ATTR_FLAG]: true,
};
const patchPropsBlackListMap = {
  transform: true,
  fill: true,
};
let moves = [];
let trackUpdate = null;
const svgTagNamesMap = (0, core_1.keyBy)(constants_1.SVG_TAG_NAMES.split(','), x => x);
const voidTagNamesMap = (0, core_1.keyBy)(constants_1.VOID_TAG_NAMES.split(','), x => x);
const createNativeElementMap = {
  [core_1.NodeType.TAG]: vNode => {
    const tagNode = vNode;
    return detectIsSvgElement(tagNode.name)
      ? document.createElementNS('http://www.w3.org/2000/svg', tagNode.name)
      : document.createElement(tagNode.name);
  },
  [core_1.NodeType.TEXT]: vNode => {
    return document.createTextNode(vNode.value);
  },
  [core_1.NodeType.COMMENT]: vNode => {
    return document.createComment(vNode.value);
  },
};
function createNativeElement(vNode) {
  return createNativeElementMap[vNode.type](vNode);
}
exports.createNativeElement = createNativeElement;
function detectIsSvgElement(tagName) {
  return Boolean(svgTagNamesMap[tagName]);
}
function detectIsVoidElement(tagName) {
  return Boolean(voidTagNamesMap[tagName]);
}
function applyRef(ref, element) {
  (0, core_1.applyRef)(ref, element);
}
function addAttributes(element, vNode) {
  if (!vNode.attrs) return;
  const attrNames = Object.keys(vNode.attrs);
  const tagElement = element;
  for (const attrName of attrNames) {
    const attrValue = vNode.attrs[attrName];
    if (attrName === core_1.ATTR_REF) {
      applyRef(attrValue, element);
      continue;
    }
    if ((0, events_1.detectIsEvent)(attrName)) {
      (0, events_1.delegateEvent)(tagElement, (0, events_1.getEventName)(attrName), attrValue);
    } else if (!(0, core_1.detectIsUndefined)(attrValue) && !attrBlackListMap[attrName]) {
      const stop = patchProperties({
        tagName: vNode.name,
        element: tagElement,
        attrValue,
        attrName,
      });
      !stop && tagElement.setAttribute(attrName, attrValue);
    }
  }
}
function updateAttributes(element, vNode, nextVNode) {
  if (!nextVNode.attrs) return;
  const attrNames = Object.keys(nextVNode.attrs);
  const tagElement = element;
  for (const attrName of attrNames) {
    const prevAttrValue = vNode.attrs[attrName];
    const nextAttrValue = nextVNode.attrs[attrName];
    if (attrName === core_1.ATTR_REF) {
      applyRef(prevAttrValue, element);
      continue;
    }
    if (!(0, core_1.detectIsUndefined)(nextAttrValue)) {
      if ((0, events_1.detectIsEvent)(attrName)) {
        prevAttrValue !== nextAttrValue &&
          (0, events_1.delegateEvent)(tagElement, (0, events_1.getEventName)(attrName), nextAttrValue);
      } else if (!attrBlackListMap[attrName] && prevAttrValue !== nextAttrValue) {
        const stop = !patchPropsBlackListMap[attrName]
          ? patchProperties({
              tagName: nextVNode.name,
              element: tagElement,
              attrValue: nextAttrValue,
              attrName,
            })
          : false;
        !stop && tagElement.setAttribute(attrName, nextAttrValue);
      }
    } else {
      tagElement.removeAttribute(attrName);
    }
  }
}
function patchProperties(options) {
  const { tagName, element, attrName, attrValue } = options;
  const fn = patchPropertiesSpecialCasesMap[tagName];
  let stop = fn ? fn(element, attrName, attrValue) : false;
  if (canSetProperty(Object.getPrototypeOf(element), attrName)) {
    element[attrName] = attrValue;
  }
  if (!stop && (0, core_1.detectIsBoolean)(attrValue)) {
    stop = !attrName.includes('-');
  }
  return stop;
}
function canSetProperty(prototype, key) {
  return prototype.hasOwnProperty(key) && Boolean(Object.getOwnPropertyDescriptor(prototype, key)?.set);
}
const patchPropertiesSpecialCasesMap = {
  input: (element, attrName, attrValue) => {
    if (attrName === 'value' && (0, core_1.detectIsBoolean)(attrValue)) {
      element.checked = attrValue;
    } else if (attrName === 'autoFocus') {
      element.autofocus = Boolean(attrValue);
    }
    return false;
  },
  textarea: (element, attrName, attrValue) => {
    if (attrName === 'value') {
      element.innerHTML = String(attrValue);
      return true;
    }
    return false;
  },
};
function commitCreation(fiber) {
  const parentFiber = (0, core_1.getFiberWithElement)(fiber.parent);
  const parentElement = parentFiber.element;
  const childNodes = parentElement.childNodes;
  if (core_1.isHydrateZone.get()) {
    const nativeElement = childNodes[fiber.eidx];
    if (
      (0, core_1.detectIsTextVirtualNode)(fiber.inst) &&
      nativeElement instanceof Text &&
      fiber.inst.value.length !== nativeElement.length
    ) {
      nativeElement.splitText(fiber.inst.value.length);
    }
    fiber.element = nativeElement;
  } else {
    if (!fiber.shadow) {
      if (childNodes.length === 0 || fiber.eidx > childNodes.length - 1) {
        !detectIsVoidElement(parentFiber.inst.name) && appendNativeElement(fiber.element, parentElement);
      } else {
        insertNativeElement(fiber.element, parentElement.childNodes[fiber.eidx], parentElement);
      }
    }
  }
  (0, core_1.detectIsTagVirtualNode)(fiber.inst) && addAttributes(fiber.element, fiber.inst);
}
function commitUpdate(fiber) {
  const element = fiber.element;
  const prevInstance = fiber.alt.inst;
  const nextInstance = fiber.inst;
  (0, core_1.detectIsPlainVirtualNode)(nextInstance)
    ? prevInstance.value !== nextInstance.value && (element.textContent = nextInstance.value)
    : updateAttributes(element, prevInstance, nextInstance);
}
function commitDeletion(fiber) {
  const parentFiber = (0, core_1.getFiberWithElement)(fiber.parent);
  if (fiber.flush) {
    parentFiber.element.textContent && (parentFiber.element.textContent = '');
    return;
  }
  (0, core_1.walkFiber)(fiber, (nextFiber, isReturn, resetIsDeepWalking, stop) => {
    if (nextFiber === fiber.next || nextFiber === fiber.parent) return stop();
    if (!isReturn && nextFiber.element) {
      !nextFiber.shadow &&
        !(0, portal_1.detectIsPortal)(nextFiber.inst) &&
        removeNativeElement(nextFiber.element, parentFiber.element);
      return resetIsDeepWalking();
    }
  });
}
function move(fiber) {
  const sourceNodes = (0, core_1.collectElements)(fiber, x => x.element);
  const sourceNode = sourceNodes[0];
  const parentElement = sourceNode.parentElement;
  const sourceFragment = new DocumentFragment();
  const elementIdx = fiber.eidx;
  let idx = 0;
  const move = () => {
    for (let i = 1; i < sourceNodes.length; i++) {
      removeNativeElement(parentElement.childNodes[elementIdx + 1], parentElement);
    }
    replaceNativeElement(sourceFragment, parentElement.childNodes[elementIdx], parentElement);
  };
  for (const node of sourceNodes) {
    insertNativeElement(document.createComment(`${elementIdx}:${idx}`), node, parentElement);
    appendNativeElement(node, sourceFragment);
    idx++;
  }
  moves.push(move);
}
const commitMap = {
  [core_1.EffectTag.C]: fiber => {
    if (fiber.element === null || (0, portal_1.detectIsPortal)(fiber.inst)) return;
    trackUpdate && trackUpdate(fiber.element);
    commitCreation(fiber);
  },
  [core_1.EffectTag.U]: fiber => {
    fiber.move && (move(fiber), (fiber.move = false));
    if (fiber.element === null || (0, portal_1.detectIsPortal)(fiber.inst)) return;
    trackUpdate && trackUpdate(fiber.element);
    commitUpdate(fiber);
  },
  [core_1.EffectTag.D]: commitDeletion,
  [core_1.EffectTag.S]: () => {},
};
function commit(fiber) {
  commitMap[fiber.tag](fiber);
}
exports.commit = commit;
function finishCommit() {
  moves.forEach(x => x());
  moves = [];
}
exports.finishCommit = finishCommit;
function setTrackUpdate(fn) {
  trackUpdate = fn;
}
exports.setTrackUpdate = setTrackUpdate;
const appendNativeElement = (element, parent) => parent.appendChild(element);
const insertNativeElement = (element, sibling, parent) => {
  parent.insertBefore(element, sibling);
};
const insertNativeElementByIndex = (element, idx, parent) => {
  parent.insertBefore(element, parent.childNodes[idx]);
};
exports.insertNativeElementByIndex = insertNativeElementByIndex;
const replaceNativeElement = (element, candidate, parent) => {
  parent.replaceChild(element, candidate);
};
const removeNativeElement = (element, parent) => parent.removeChild(element);
//# sourceMappingURL=dom.js.map
